#pragma once
#include "types.h"
#include "enumerations.h"

/*
 *  Structs by Poiuy_qwert and me. ++updated with the BWAPI source
 */

#pragma pack(1)
struct POSITION {
  u16 x;
  u16 y;
};

struct UNIT;

//Based on http://code.google.com/p/bwapi/source/browse/branches/bwapi4/bwapi/BWAPI/Source/BW/Target.h
struct TARGET {
  POSITION  pt;
  UNIT*     unit;
};

struct UNITDEATHS {
  u32        player[12];
};

struct SPRITE;

struct ORDER;

//Based on http://code.google.com/p/bwapi/source/browse/branches/bwapi4/bwapi/BWAPI/Source/BW/CImage.h
struct IMAGE {
  IMAGE*      previous;
  IMAGE*      next;
  u16         imageId;
  u8          paletteType;          //++Drawing function (values are the same as DataEdit)
  u8          direction;
  u16         flags;
              /*  0x0001  - Redraw
                  0x0002  - Don't update x?
                  0x0004  - Don't update y?
                  0x0008  -
                  0x0010  -
                  0x0020  -
                  0x0040  - Hidden/Invisible (don't draw)
                  0x0080  -
                  0x0100  -
                  0x0200  -
                  0x0400  -
                  0x0800  -
                  0x1000  -
                  0x2000  -
                  0x4000  -
                  0x8000  -
              */
  u8          horizontalOffset;
  u8          verticalOffset;
  u16         iscriptHeaderOffset;
  u16         iscriptOffset;
  UNK         unknown2[2];
  u8          animation;
  u8          wait;
  u16         frameSet;
  u16         frameIndex;
  POSITION    mapPosition;
  POSITION    screenPosition;
  UNK         unknown3[8];
  u32         grpOffset;
  void*       coloringData;        //?
  void*       renderFunction;
  void*       updateFunction;
  SPRITE*     parentSprite;
};

C_ASSERT(sizeof(IMAGE) == 64);

//Based on http://code.google.com/p/bwapi/source/browse/trunk/bwapi/BWAPI/Source/BW/Sprite.h
//Also see http://code.google.com/p/bwapi/source/browse/branches/bwapi4/bwapi/BWAPI/Source/BW/CSprite.h
struct SPRITE {
  SPRITE*   previous;
  SPRITE*   next;
  u16       id;               ///< Sprite ID in sprites.dat
  u8        playerId;         ///< Officially "creator". Also controls team color?
  u8        selectionIndex;   ///< A value between 0 and 11, inclusive. Index in the selection area at bottom of screen.
  u8        visibilityFlags;
  u8        elevationLevel;
  u8        flags;
            /*  0x01  Draw selection circle.
                0x02
                0x04
                0x08  Selected.
                0x10
                0x20  Hidden
                0x40  Burrowed
                0x80  Iscript unbreakable code section.
            */
  u8        selectionTimer;
  u32       index;
  POSITION  position;
  IMAGE*    mainGraphic;
  IMAGE*    underlay;         ///< selection circle
  IMAGE*    overlay;          ///< shadow
};

C_ASSERT(sizeof(SPRITE) == 36);

struct TEXT {
  char       textdisplay[218][13];  //[last is text at bottom; ex: "Cheat enabled."]
  UNK        unknown1;
  UNK        unknown2;
  u8         color[13];
  UNK        unknown3[3];  //[last is text at bottom; ex: "Cheat enabled."]
  u8         timer[13];
};

struct TILE {
      u16 index;
      u8 buildability; /**< 8th bit should sign not buildable. */
      u8 groundHeight; /**< Ground Height(4 lower bits) - Deprecated? Some values are incorrect. */
      u16 leftEdge;
      u16 topEdge;
      u16 rightEdge;
      u16 buttomEdge;
      u16 _1;
      u16 _2; /**<  Unknown - Edge piece has rows above it. (Recently noticed; not fully understood.)
                    o 1 = Basic edge piece.
                    o 2 = Right edge piece.
                    o 3 = Left edge piece.*/
      u16 _3;
      u16 _4;
      u16 miniTile[16]; /** MegaTile References (VF4/VX4) */
};

struct DOODAD {
      u16 index;
      u8 buildability; /**< 8th bit should sign not buildable. */
      u8 groundHeightAndOverlayFlags; /**< Ground Height(4 lower bits) - Deprecated? Some values are incorrect.
                       * Overlay Flags:
                       * o 0x0 - None
                       * o 0x1 - Sprites.dat Reference
                       * o 0x2 - Units.dat Reference (unit sprite)
                       * o 0x4 - Overlay is Flipped
                       */
      u16 overlayID;
      u16 _1;
      u16 doodatGrupString;
      u16 _2;
      u16 dddataBinIndex;
      u16 doodatHeight;
      u16 doodatWidth;
      u16 _3;
      u16 miniTile[16]; /** MegaTile References (VF4/VX4) */
};

//---- Taken from locations.cpp ----//

struct MAPDIMENSIONS {
  u16        width;
  u16        height;
};

struct LOCATION {
  u32       topLeftX;
  u32       topLeftY;
  u32       bottomRightX;
  u32       bottomRightY;
  u16       stringId;
  u16       flags;
};

//---- Taken from player.cpp ----//

struct PLAYER {
  u32             id;
  u32           actions; // unused; FF FF FF FF if not a human player
  u8            type;
  Player::Race  race;
  u8            force;
  char          name[25];
};

//---- Taken from buttons.cpp ----//

//typedef struct UNIT UNIT;

typedef BUTTON_STATE::Enum (__fastcall *REQ_FUNC)(u32, u32, UNIT*);
//(u32 reqVar, u32 playerId, UNIT* selected);
typedef void (__fastcall *ACT_FUNC)(u32,u32);
//(u32 actVar, u32 shiftClick);

struct BUTTON {
  u16        position;
  u16        iconID;
  REQ_FUNC*    reqFunc;
  ACT_FUNC*    actFunc;
  u16        reqVar;
  u16        actVar;
  u16        reqStringID;
  u16        actStringID;
};

struct BUTTON_SET {
  u32        buttonsInSet;
  BUTTON*      firstButton;
  u32        connectedUnit;
};

//---- Taken from resources.cpp ----//

struct RESOURCES {
  int        playerMin[12];
  int        playerGas[12];
  int        cumulativeGas[12];
  int        cumulativeMinerals[12];
};

//---- Taken from triggers.h ----//
struct CONDITION {
  u32        location;
  u32        player;
  u32        number;
  u16        unit;
  u8        comparison;
  u8        condition;
  u8        type;
  u8        flags;
};

typedef BOOL (__fastcall *ConditionPointer)(CONDITION*);

struct ACTION {
  u32        location;
  u32        string;
  u32        wavString;
  u32        time;
  u32        player;
  u32        number;
  u16        unit;
  u8        action;
  u8        number2;
  u8        flags;
};

typedef BOOL (__fastcall *ActionPointer)(ACTION*);

#pragma pack()
