/// This header file contains the structures and functions that allow you to
/// change the maximum sight range limit in StarCraft.

/// To use, add the following code to the gameOn() function in hooks.cpp:
///		sightRangeLimitExtender.setMaxRange(24);    //Sets max sight range to 24

#pragma once
#include "offsets.h"
#include "../definitions.h"
#include "../tools.h"
#pragma pack(1)

//Warning: Do NOT attempt to create an instance of this structure by yourself!
struct SightRangeLimitExtender {
	struct SightStruct {
		u32 tileSightWidth;
		u32 tileSightHeight;
		u32 unknown1;
		u32 unknown2;
		u32 unknown3;
		u32 unknown4;
		u32 unknown5;
		SightStruct() : unknown1(3), unknown2(3), unknown3(0), unknown4(0), unknown5(0) {}
	} *sightData;

	void setMaxRange(unsigned int i) {
		delete [] sightData;
		sightData = new SightStruct[++i];   //Since sightData = [sight_0, sight_1, ... , sight_i]
		for (unsigned int x = 0; x < i; ++x) {
			sightData[x].tileSightWidth = sightData[x].tileSightHeight = x * 2 + 3;
		}
		memoryPatch4((void*)VERSION::SightStructArrayRef, (u32)sightData);

		void* p = sightData;  //To fix a weird bug where VC++ "optimizes" MOV ECX, sightData to MOV ECX, 0
		__asm {
			PUSHAD
			MOV EAX, VERSION::MapTileWidth
			MOVZX EAX, WORD PTR DS:[EAX]
			PUSH EAX
			MOV EAX, i
			MOV ECX, p
			CALL [VERSION::InitializeSightValues]
			POPAD
		}
	}

  //Called when the plugin is loaded
	SightRangeLimitExtender() : sightData(NULL) {}
  //Called when the plugin is unloaded (when StarCraft.exe is closed)
	~SightRangeLimitExtender() { delete [] sightData; }
};

static SightRangeLimitExtender sightRangeLimitExtender;

#pragma pack()
