#pragma once
#include "structures.h"

#pragma pack(1)

//Comments marked with ++ added by RavenWolf
struct UNIT {
  UNIT*     previous;
  UNIT*     next;
  u32       healthPoints;
  SPRITE*   sprite;
  TARGET    moveTarget;         ///< The position or unit to move to. It is NOT an order target.
  POSITION  nextMovementWaypoint; ///< The next way point in the path the unit is following to get to its destination.
                                  ///< Equal to moveToPos for air units since they don't need to navigate around buildings.
  POSITION  nextTargetWaypoint; ///< The desired position
  u8        movementFlags;
  u8        currentDirection1;
  u8        flingyTurnSpeed;    //AKA flingy turn radius
  u8        velocityDirection1;
  u16       flingyId;
  UNK unknown2;
  u8        flingyMovementType;
  POSITION  position;
  u32       xHalt;
  u32       yHalt;
  u32       flingyTopSpeed;     //top speed
  u32       current_speed1;     //++current speed
  u32       current_speed2;     //++not sure if there are any diference with this 2 variables, they always have the same value
  UNK unknown3[8];
  u16       flingyAcceleration;
  u8        currentDirection2;  //?
  u8        velocityDirection2; //?
  u8        playerId;
  u8        mainOrderId;
  u8        mainOrderState;
  u8        orderSignal;        /*  0x01  Update building graphic/state
                                    0x02  Casting spell
                                    0x04  Reset collision? Always enabled for hallucination...
                                    0x10  Lift/Land state
                                */
  u16       orderUnitType;
  UNK unknown5[2];
  u8        mainOrderTimer;
  u8        groundWeaponCooldown;
  u8        airWeaponCooldown;
  u8        spellCooldown;
  TARGET    orderTarget;
  u32       shields;
  u16       unitId;
  UNK unknown6[2];
  UNIT*     playerPrevious;
  UNIT*     playerNext;
  UNIT*     subunit;
  ORDER*    orderQueueHead;
  ORDER*    orderQueueTail;
  UNIT*     autoTargetUnit;
  UNIT*     connectedUnit;      // ++Larva store a reference to the hatchery/Lair/Hive who created it
                                // ++units inside transports store a reference to the transport
                                // addons store a reference to its parent building (but not the other way around)
  u8        orderQueueCount;
  u8        orderQueueTimer;    //?
  UNK unknown8;
  u8        attackNotifyTimer;  // Prevent "Your forces are under attack." on every attack
  u16       displayedUnitId;    //++used by zerg buildings while morphing
  u8        lastEventTimer;     // countdown that stops being recent when it hits 0
  u8        lastEventColor;     // 17 = was completed (train, morph), 174 = was attacked

  UNK unknown9[2];
  u8        rankIncrease;
  u8        killCount;
  u8        lastAttackingPlayer;// the player that last attacked this unit
  u8        secondaryOrderTimer;
  u8        aiActionFlag;       // Used internally by the AI
  u8        userActionFlags;    // some flags that change when the user interacts with the unit
                                // 2 = issued an order, 3 = interrupted an order, 4 = self destructing
  u16       currentButtonSet;   // The u16 is a guess, used to be u8
  u8        isCloaked;          // 1 for cloaked
  u8        movementState;      // A value based on conditions related to pathing, see Path.h for info

  u16       buildQueue[5];    //++train queue for buildings and carrier/reaver (stores the unitid) / zerg eggs stores the id of the morphing unit in queue[0]
  u16       energy;
  u8        buildQueueSlot;     //< Index of active unit in #buildQueue.
  u8        targetOrderSpecial; //< A byte used to determine the target ID for the unit
  u8        secondaryOrderId;
  u8        buildingOverlayState; // 0 means the building has the largest amount of fire/blood
  u16       buildRepairHPGain;  //++hp ammount increased per frame when a unit/building its constructed/repaired
  UNK unknown14[2];
  u16       remainingBuildTime;
  u16       previousHp;         // The HP of the unit before it changed (example Drone->Hatchery, the Drone's HP will be stored here)
  u16       loadedUnitIndex[8];

  union {   //0x0C0
    struct {
      u8 spiderMineCount; // 0
    } vulture;

    struct {
      UNIT*   inHangerChild;    // 0   first child inside the hanger
      UNIT*   outHangerChild;   // 4   first child outside the hanger
      u8      inHangerCount;    // 8   number inside the hanger
      u8      outHangerCount;   // 9   number outside the hanger
    } carrier; // also applies to reaver

    struct {
      UNIT*   parent;   // 0
      UNIT*   prev;     // 4
      UNIT*   next;     // 8
      u8      inHanger;  // C
    } interceptor;  // also applies to scarab

    struct {
      u32 _unknown_00;
      u32 _unknown_04;
      u32 flagSpawnFrame; // flag beacons, the frame that the flag will spawn
    } beacon;

    struct {
      UNIT*   addon;                // 0
      u16     addonBuildType;       // 4
      u16     upgradeResearchTime;  // 6
      u8      techType;             // 8
      u8      upgradeType;          // 9
      u8      larvaTimer;           // A
      u8      landingTimer;         // B
      u8      creepTimer;           // C
      u8      upgradeLevel;         // D
      u16     _padding_0E;          // E    // possible alignment
      union {
        struct {
          u16       resourceCount;  // amount of resources
          u8        resourceIscript;
          u8        gatherQueueCount;
          UNIT*     nextGatherer;   // 14  // pointer to the next workerunit waiting in line to gather
          u8        resourceGroup;  // 18
          u8        resourceBelongsToAI;
        } resource;  /** When the unit is resource container */
        struct { UNIT* exit; } nydus; /** connected nydius canal */
        struct { SPRITE* nukeDot; } ghost;
        struct { SPRITE* pylonAura; } pylon;
        struct{
          UNIT* nuke;  // attached nuke
          u32 hasNuke;     // 14
        } silo;
        struct{
          u16 harvestValue1;
          u16 harvestValue2;
          u16 _unknown_14;    // 14
          u16 harvestValue3;
        } hatchery; // wtf???
        struct {
          u16  powerupX;
          u16  powerupY;
        } powerup;
      };
    } building;

    struct {
      UNIT* powerup;                // 0
      u16   targetResourceX;        // 4
      u16   targetResourceY;
      UNIT* targetResourceUnit;     // 8
      u16   repairResourceLossTimer;// C
      u8    isCarryingSomething;    // E
      u8    resourceCarryCount;     // F
      UNIT* harvestTarget;          // 10
      UNIT* prevHarvestUnit;        // 14   // When there is a gather conflict
      UNIT* nextHarvestUnit;        // 18
    } worker;
  };

  u32       status;             // See UnitStatus::Enum for more info
  u8        resourceType;
  u8        wireframeRandomizer;
  u8        secondaryOrderState;
  u8        recentOrderTimer;   // Counts down from 15 to 0 when most orders are given, or when the unit moves after reaching a patrol location
  s32       visibilityStatus;   // Flags specifying which players can detect this unit (cloaked/burrowed)
  UNK unknown17[4];
  UNIT*     currentBuildUnit;
  UNIT*     previousBurrowedUnit;
  UNIT*     nextBurrowedUnit;
  union {
    TARGET rally;               // This is not present in BWAPI's CUnit.h, but it seems pretty obvious that it is a TARGET struct.
    struct {
      UNIT* previousPsiProvider;
      UNIT* nextPsiProvider;
    } pylon;
  };
  u32       path;               //?
  u8        pathingCollisionInterval;  // unknown
  u8        pathingFlags;       // 0x01 = uses pathing; 0x02 = ?; 0x04 = ?
  u8        unused_0x106;
  u8        isBeingHealed;
  u16       contours1XUnknown;  //?
  u16       contours1YUnknown;  //?
  u16       contours2XUnknown;  //?
  u16       contours2YUnknown;  //?
  u16       removeTimer;        // Hallucination, Dark Swarm, Disruption Web, Broodling
  u16       defensiveMatrixHp;
  u8        defensiveMatrixTimer;
  u8        stimTimer;
  u8        ensnareTimer;
  u8        lockdownTimer;
  u8        irradiateTimer;
  u8        statisTimer;
  u8        plagueTimer;
  u8        isUnderStorm;       // Used to tell if a unit is under psi storm
  UNIT*     irradiatedBy;
  u8        irradiatePlayerId;
  u8        parasiteFlags;      // Each bit corrisponds to the player who has parasited this unit*/
  u8        cycleCounter;       // counts/cycles up from 0 to 7 (inclusive).
  u8        isBlind;
  u8        maelstromCounter;
  u8        unusedTimer;        // ?? Might be afterburner timer or ultralisk roar timer

/*  #ifdef VERSION v1151
    u8        unknownUnreferenced;    //++not available in 1.16.1
  #endif*/
  u8        acidSporeCount;
  u8        acidSporeTime[9];
  u16       offsetIndex3by3;
  UNK unknown21[6];
  u16       airStrength;
  u16       groundStrength;
  u32       tableIdUnknown[4];        //?
  u8        repulse1Unknown;        //?
  u8        repulseAngle;          // updated only when air unit is being pushed
  u8        driftPosX;
  u8        driftPosY;
};

C_ASSERT(sizeof(UNIT) == 336);


/*
struct WEAPON{
  WEAPON*      previousWeapon;
  WEAPON*      nextWeapon;
  UNK unknown1[4];
  SPRITE*      sprite;
  u16        targetGroundX;
  u16        targetGroundY;
  UNIT*      targetUnit;
  u8        weaponId;
  UNK unknown2[2];
  u8        bounceCount;
  UNIT*      attackingUnit;
};*/

  // See Unit.h and compare this entire structure with the first part of the unit structure
  // I'm not sure how much of it is the same so I didn't bother updating this.
  struct BULLET {
// v-- POSSIBLE SHARED BULLET/UNIT STRUCTURE BEGIN
    /*0x00*/ BULLET*      previous;
    /*0x04*/ BULLET*      next;
    /*0x08*/ u32          unknown_0x08; //usually holds 1 (exists?)
    /*0x0C*/ SPRITE*      sprite; // sprite pointer
    /*0x10*/ TARGET       moveTarget; //address? animation related? changes a during the same bullet
    /*0x18*/ POSITION     targetPosition;
    /*0x1C*/ POSITION     targetPosition2; //usually same as targetPosition
    /*0x20*/ u8           movementFlags; //usually holds 0x09 for probes, 0x0b for dragoons, 0x11 for mutas
    /*0x21*/ u8           direction1; //not sure how this and direction2 differ from currentDirection and velocityDirection
    /*0x22*/ u8           flingyTurnRadius;
    /*0x23*/ u8           direction2;
    /*0x24*/ u16          type;
    /*0x26*/ u8           unknown_0x26;
    /*0x27*/ u8           flingyMoveControl;
    /*0x28*/ POSITION     position;
    /*0x2C*/ POINT        halt; // (xHalt >> 4) == xHalt
    /*0x34*/ s32          flingyTopSpeed;
    /*0x38*/ s32          unknown_0x38; //increasing counter or timer
    /*0x3C*/ s32          unknown_0x3C; //increasing counter or timer, often the same value as unknown_0x38
    /*0x40*/ POINT        current_speed;
    /*0x48*/ u16          flingyAcceleration;
    /*0x4A*/ u8           currentDirection;
    /*0x4B*/ u8           velocityDirection; //seems to always be the same as currentDirection
    /*0x4C*/ u8           unknown_0x4C;
    /*0x4D*/ u8           behaviourTypeInternal; // 0 = none
                                                 // 1 = fly
                                                 // 2 = fly & follow target
                                                 // 3 = bounce
                                                 // 4 = persist on site
                                                 // 5 = appear on target/instant
                                                 // 6 = attack target area
    /*0x4E*/ u16          unknown_0x4E;
    /*0x50*/ u16          someUnitType; // usually holds UnitType::None (Unknown unit type stuff?)
    /*0x52*/ u16          unknown_0x52; // possibly padding
    /*0x54*/ s32          unknown_0x54; // usually holds 0
    /*0x58*/ TARGET       attackTarget;
// ^-- POSSIBLE SHARED BULLET/UNIT STRUCTURE END

    /*0x60*/ u8           weaponType;
    /*0x61*/ u8           time_remaining;
    /*0x62*/ u8           hitFlags;           // 0x01 = miss target; 0x02 = no damage (hallucination)
    /*0x63*/ u8           remainingBounces;   // Bounce count for mutalisk weapon (official name: ubChain)
    /*0x64*/ UNIT*        sourceUnit;
    /*0x68*/ UNIT*        nextBounceUnit;     // Next unit for bounce, temporary
    /*0x6C*/ u32          unknown_0x6C;       // some sort of cyclic burst counter (used by valkyries, goliath air attack)
  };

 C_ASSERT(sizeof(BULLET) == 112);

#pragma pack()
