/// This is where the magic happens; program your plug-in's core behavior here.
/// Note that the hooks must be enabled in initialize.h in order to work!

#include "SCBW/scbwdata.h"
#include "SCBW/enumerations.h"
#include "SCBW/api.h"
#include "SCBW/ExtendSightLimit.h"
#include "hooks.h"
#include <cstdio>

/******** Core hook functions ********/

bool firstRun = true;

bool nextFrame() {
	if (!*IS_GAME_PAUSED) { //If the game is not paused

		if (firstRun) {
			scbw::printText("Hello, world!");
			firstRun = false;
		}

		// Loop through the unit table.
    // Warning: There is no guarantee that the current unit is actually a unit
    // rather than an unused space in memory.
		for (int i = 0; i < UNIT_ARRAY_LENGTH; ++i) {
			UNIT* unit = &unitTable[i];
      //Write your code here
		}

    // Alternative looping method
    // Guarantees that [unit] points to an actual unit.
    //for (UNIT* unit = *firstVisibleUnit; unit; unit = unit->next) {
    //  //Write your code here
    //}

		// Loop through the bullet table.
    // Warning: There is no guarantee that the current bullet is actually a
    // bullet rather than an unused space in memory
		for (int i = 0; i < BULLET_ARRAY_LENGTH; ++i) {
			BULLET* bullet = &bulletTable[i];
      //Write your code here
		}

    // Alternative looping method
    // Guarantees that [bullet] points to an actual bullet.
    //for (BULLET* bullet = *firstBullet; bullet; bullet = bullet->next) {
    //  //Write your code here
    //}
	}
	return true;
}

bool gameOn() {
	firstRun = true;
	//Uncomment to use this feature
	//sightRangeLimitExtender.setMaxRange(11);	//Change 11 to whatever value you wish
	return true;
}

bool gameEnd() {
	return true;
}


/******** Utility hook functions (enable in initialize.h) ********/

/// Tells the game the HP cost of Stim Packs for the given unit.
/// Remember: HP in memory is equal to (HP displayed) * 256!
u32 getStimPacksHpCost(const UNIT* unit) {
	//Default StarCraft behavior
	//return 2560;
	return 7 * 256;
}

/// Handles the job of regenerating the unit's hit points.
void regenerateUnitHp(UNIT* unit) {
	using Unit::BaseProperty;
	using Unit::MaxHitPoints;
	using scbw::setUnitHp;

	//Default StarCraft behavior
	if (BaseProperty[unit->unitId] & UnitProperty::RegeneratesHP) {
		u32 previousHp = unit->healthPoints;
		if (previousHp && previousHp != MaxHitPoints[unit->unitId])
			setUnitHp(unit, previousHp + *HP_REGEN_AMOUNT);
	}
}

/// Tells the game how much energy a unit should regenerate per each tick.
u16 energyRegenAmount(const UNIT* unit) {
	//Default StarCraft behavior
	return *ENERGY_REGEN_AMOUNT;
}

/// Decides whether the given unit can be Consumed.
bool canConsumeUnit(const UNIT* unit, u8 consumingPlayer) {
	using Unit::BaseProperty;
	using Unit::GroupFlags;

	//Default StarCraft behavior
	return !(BaseProperty[unit->unitId] & UnitProperty::Building)	//Cannot consume buildings
		&& unit->playerId == consumingPlayer				//Must be owned by the current player
		&& GroupFlags[unit->unitId] & UnitGroup::Zerg		//Must be a Zerg unit
		&& unit->unitId != UnitId::larva						//Cannot consume larvae
	;
}

/// Tells the game how much energy should be given when this unit is consumed.
/// Remember: energy in memory is equal to (energy displayed) * 256!
u32 consumeEnergyGain(const UNIT* unit) {
	//Default StarCraft behavior
	return *CONSUME_ENERGY_GAIN;
}

/// Handles the job of applying the actual effects of Plague.
void doPlagueDamage(UNIT* unit) {
	using Weapon::DamageAmount;
	using Unit::ShieldsEnabled;

	//Default StarCraft behavior
	if (!(unit->status & UnitProperty::Invincible)) {
		u32 damage = DamageAmount[60] / 76;		//WEAPONS.DAT entry: [60] Plague
		if (damage < unit->healthPoints)
			scbw::damageUnitHp(unit, damage, NULL, unit->lastAttackingPlayer);
	}
}

/// Tells the game how much plasma shields the unit can recharge up to.
u32 maxRechargeShields(const UNIT* unit) {
	using Unit::MaxShieldPoints;
	//Default StarCraft behavior
	return MaxShieldPoints[unit->unitId];
}

/// Handles the job of regenerating the unit's plasma shields.
///
/// @return		A boolean value indicating whether the unit's shields were changed by this function.
bool regenerateShields(UNIT* unit) {
	using Unit::MaxShieldPoints;

	//Default StarCraft behavior
	u32 maxShields = MaxShieldPoints[unit->unitId] << 8;
	if (unit->shields == maxShields) return false;	//Shields are unchanged

	unit->shields += *SHIELD_REGEN_AMOUNT;
	if (unit->shields > maxShields) unit->shields = maxShields;
	return true;	//Shields are changed
}

/// Calculates the unit's modified movement speed, factoring in upgrades and status effects.
///
/// @return		The modified speed value.
u32 modifyUnitSpeed(const UNIT* unit, u32 baseSpeed) {
	//Default StarCraft behavior
	u32 speed = baseSpeed;
	u32 speedModifier = (unit->stimTimer ? 1 : 0) - (unit->ensnareTimer ? 1 : 0)
		+ (unit->status & UnitStatus::SpeedUpgrade ? 1 : 0);
	if (speedModifier > 0) {
		if (unit->unitId == UnitId::scout || unit->unitId == UnitId::Hero_Mojo || unit->unitId == UnitId::Hero_Artanis)
		speed = 1707;
		else {
			speed += speed >> 1;
			speed = speed > 853 ? speed : 853;
		}
	}
	else if (speedModifier < 0)
		speed >>= 1;
	return speed;
}

/// Calculates the unit's acceleration, factoring in upgrades and status effects.
///
/// @return		The modified acceleration value.
u32 modifyUnitAcceleration(const UNIT* unit) {
	//Default StarCraft behavior
	u32 acceleration = Flingy::Acceleration[Unit::Graphic[unit->unitId]];
	u32 modifier = (unit->stimTimer ? 1 : 0) - (unit->ensnareTimer ? 1 : 0)
		+ (unit->status & UnitStatus::SpeedUpgrade ? 1 : 0);
	if (modifier > 0)
		acceleration *= 2;
	else if (modifier < 0)
		acceleration -= acceleration >> 2;
	return acceleration;
}

/// Calculates the unit's turn speed, factoring in upgrades and status effects.
///
/// @return		The modified turning speed value.
u32 modifyUnitTurnSpeed(const UNIT* unit) {
	//Default StarCraft behavior
	u32 turnSpeed = Flingy::TurnSpeed[Unit::Graphic[unit->unitId]];
	u32 modifier = (unit->stimTimer ? 1 : 0) - (unit->ensnareTimer ? 1 : 0)
		+ (unit->status & UnitStatus::SpeedUpgrade ? 1 : 0);
	if (modifier > 0)
		turnSpeed *= 2;
	else if (modifier < 0)
		turnSpeed -= turnSpeed >> 2;
	return turnSpeed;
}

/// Calculates the unit's weapon cooldown, factoring in upgrades and status effects.
///
/// @return		The modified cooldown value.
u32 modifyWeaponCooldown(const UNIT* unit, u8 weaponId) {
	//Default StarCraft behavior
	u32 cooldown = Weapon::Cooldown[weaponId];

	if (unit->acidSporeCount) {
		u32 increaseAmt = cooldown >> 3;
		if (increaseAmt < 3) increaseAmt = 3;
		cooldown += increaseAmt * unit->acidSporeCount;
	}

	u32 cooldownModifier = (unit->stimTimer ? 1 : 0) - (unit->ensnareTimer ? 1 : 0)
		+ (unit->status & UnitStatus::CooldownUpgrade ? 1 : 0);
	if (cooldownModifier > 0)
		cooldown >>= 1;
	else if (cooldownModifier < 0)
		cooldown += cooldown >> 2;

	if (cooldown > 250) cooldown = 250;
	else if (cooldown < 5) cooldown = 5;

	return cooldown;
}

/// Orders newly-produced units to rally, based upon the properties of the
/// building that produced it.
///
/// @param  unit      The unit that needs to receive rally orders.
/// @param  factory   The unit (building) that created the given unit.
void orderNewUnitToRally(UNIT* unit, UNIT* factory) {
  //Auto-harvest, auto-enter transport
  using Unit::BaseProperty;
  using scbw::orderToUnit;
  using scbw::orderToPos;

  if (factory->rally.unit == factory || !(factory->rally.pt.x))
    return;

  UNIT* targetUnit = factory->rally.unit;

  if (targetUnit) {
    //Check if the target can be harvested from
    if (BaseProperty[targetUnit->unitId] & UnitProperty::ResourceContainer
        && (targetUnit->unitId >= 176 && targetUnit->unitId <= 178  //Is a mineral patch (IDs 176-178)
            || targetUnit->playerId == unit->playerId)              //Is a gas building owned by the player
        ) {
      if (BaseProperty[unit->unitId] & UnitProperty::Worker) {
        orderToUnit(unit, OrderId::Harvest1, targetUnit);
        return;
      }
      if (factory->unitId >= 131 && factory->unitId <= 133) {       //Is a Hatchery, Lair or Hive (IDs 131-133)
        orderToPos(unit, OrderId::Move, factory->rally.pt.x, factory->rally.pt.y);
        return;
      }
      //Otherwise, continue to the next check
    }

    //Check if the target can be entered
    if (scbw::canBeEnteredBy(targetUnit, unit)) {
      orderToUnit(unit, OrderId::EnterTransport, targetUnit);
      return;
    }
    if (targetUnit->playerId == unit->playerId) {
      orderToUnit(unit, OrderId::Follow, targetUnit);
      return;
    }
    //Cannot use targetUnit, so move on
  }

  orderToPos(unit, OrderId::Move, factory->rally.pt.x, factory->rally.pt.y);
}
